#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"


/* no. of processors in the x and y directions */
#define nxp 2
#define nyp 2

/* no. of grid points in the x and y directions overall */
#define nx 48
#define ny 48

#define ndim 2
/* size of data block in each processor is nx/nxp and ny/nyp   */
#define nxblock 24 
#define nyblock 24

/* max index of arrays in each processor, includes guard ring */
#define nxlocal 25
#define nylocal 25


int main(int argc, char *argv[])
{
  
  MPI_Status status;

  int rank, value, size, i, j, k, iters=0;
  int my_grid_rank;
    
  double toterr=0.0, myerr=0.0;
    
    /* variables for 2D communication   */

  MPI_Comm   grid_comm;
  MPI_Datatype  rowtype;
  MPI_Datatype  coltype;
    
  int coord[ndim];
  int dims[2];
  int periods[2];

  int rankx, ranky;
  int reorder;
  int coordinates[2];

  int nbrleft, nbrright, nbrbottom, nbrtop;
    
    /* arrays for field values in each processor
     array u is the field values, unew is a temp, and f has the
     right hand sides of the poisson equation
      mask is true on the interior and false on the boundary */

  double u[nxlocal][nylocal];
  double unew[nxlocal][nylocal];
  double f[nxlocal][nylocal];
  int mask[nxlocal][nylocal];

    /* dx and dy are the interval sizes in the two dimensions
     the equations are being solved over interval from 0 to 1 in
     both dimensions  */

  double dx=1.0/nx;
  double dy=1.0/ny;
    
  dims[0] = nxp;
  dims[1] = nyp;
    
    /* no circular shift */
  periods[0] = 0;
  periods[1] = 0;
    
  reorder = 1;
    
  MPI_Init( &argc, &argv );


  MPI_Comm_rank( MPI_COMM_WORLD, &rank );
  MPI_Comm_size( MPI_COMM_WORLD, &size );

  MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder,
		  &grid_comm);

  MPI_Comm_rank(grid_comm, &my_grid_rank);
  MPI_Cart_coords(grid_comm, my_grid_rank, 2, coordinates);

  MPI_Cart_shift(grid_comm, 1, 1, &nbrleft, &nbrright);
  MPI_Cart_shift(grid_comm, 0, 1, &nbrbottom, &nbrtop);

  rankx = coordinates[0];
  ranky = coordinates[1];

  if (rank==0){ printf ("rankx %d ranky %d \n", rankx, ranky);}

#if 1
  MPI_Type_vector(nxblock, 1, nylocal+1, MPI_DOUBLE, &coltype);
  MPI_Type_contiguous(nyblock, MPI_DOUBLE, &rowtype);
#endif

  MPI_Type_commit(&coltype);
  MPI_Type_commit(&rowtype);
    
/* initialize u, unew, f and mask arrays */
  for (k=0; k < nylocal+1; k++) 
    {
      for (i=0; i < nxlocal+1; i++) 
	{
	  
	  f[i][k] = - 2.0 *((dx *(i + coordinates[0] * nxblock)) *
			    (dx *(i + coordinates[0] * nxblock)))
	    + (2.0 * dx * (i + coordinates[0] * nxblock));

	  if (rank == 0)
	    printf(" i = %d  and k = %d and f = %lf \n", i, k, f[i][k]);

	  u[i][k] = 0.0;
	  unew[i][k] = 0.0;
	  mask[i][k] = 1;
	}
    }
  if (rank == 0)
    printf(".........final k = %d,   final i =%i\n", k, i);
  
  /* change mask to be false on the boundaries of the big array */
  if (rankx == 0)
    { for (i=0; i < nylocal; i++) { mask[1][i] = 0; }
    }
  if (rankx == nxp-1)
    { for (i=0; i < nylocal; i++) { mask[nxlocal-1][i] = 0; }}
  if (ranky == 0)
    { for (i=0; i < nxlocal; i++) { mask[i][1] = 0; }}
  if (rankx == nyp-1)
    { for (i=0; i < nxlocal; i++) { mask[i][nylocal-1] = 0;} }

  /* Processors 0 and 1 exchange, 2 and 3 exchange, etc.  Then
     1 and 2 exchange, 3 and 4, etc.  The formula for this is
     if (even) exchng up else down
     if (odd)  exchng up else down
  */

  
  do {
    iters++;
    if (rank == 0)
      printf("My error %lf at iteration %d\n", myerr, iters);
    
    /* communicate edge values with neighbors */
    if (grid_comm != MPI_COMM_NULL) {
      MPI_Sendrecv( &(u[1][1]), 1, coltype, nbrleft, 0, 
		    &(u[1][nylocal]), 1, coltype, nbrright, 0, 
		    grid_comm, &status );

      MPI_Sendrecv( &(u[1][nylocal-1]), 1, coltype, nbrright, 1,
		    &(u[1][0]), 1, coltype, nbrleft, 1, 
		    grid_comm, &status );

      MPI_Sendrecv( &(u[1][1]), 1, rowtype, nbrtop, 2, 
		    &(u[nxlocal][1]), 1, rowtype, nbrbottom, 2, 
		    grid_comm, &status );

      MPI_Sendrecv( &(u[nxlocal-1][1]), 1, rowtype, nbrbottom, 3,
		    &(u[0][1]), 1, rowtype, nbrtop, 3, 
		    grid_comm, &status );
    }
    

    /* compute one jacobi iteration of new field values for all
       interior points of the array. */
    myerr= 0.0;
    for (j=1; j <= nylocal-1; j++)
      {
	for (i=1; i <= nxlocal-1; i++)
	  { 
	    if (mask[i,j])
	      {
		unew[i][j] = (u[i-1][j] + u[i+1][j] +
			      u[i][j-1] + u[i][j+1] +
			      f[i][j])/4.0;
		if (fabs( unew[i][j] - u[i][j]) > myerr)
		  myerr = fabs(unew[i][j] - u[i][j]);
	      }
	  }
      }


    MPI_Allreduce( &myerr, &toterr, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

    for (j=1; j <= nylocal-1; j++)
      {
	for (i=1; i <= nxlocal-1; i++)
	  {
	    u[i][j]=unew[i][j];
	  }
      }
    if (rank == 0)
      printf ("The total error (toterr) = %lf\n", toterr);
    
      
  } while(toterr > 1.0e-1); 
    
  if (my_grid_rank == 0) {
    printf( "! resulting %lf error\n", toterr );
    printf( "| no. of iterations %d \n", iters );
  }
  MPI_Finalize( );
  return 0;
}






